/*
 * Copyright 2016 Aaron L. Zeng <a2z@mit.edu>, Yue Qiu <yqiu2@wellesley.edu>, Emily Armstrong<earmstro24@gmail.com>
 */

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "./quadtree.h"
#include <cilk/cilk.h>
#include <cilk/reducer.h>

#define CHILD_COUNT 4
#define LINE_THRESHOLD 3

// Enum is defined this way so that NW=0, NE=1, etc., and names can be used
// instead of indices.
enum line_category {
  NW,
  NE,
  SW,
  SE,
  SELF,
};

typedef enum line_category line_category_t;

static void assert_invariant(const quadtree_t *qt) {
  assert(qt->num_lines >= 0);
  if (qt->children != NULL) {
    for (size_t i = 0; i < CHILD_COUNT; i++) {
      assert_invariant(&qt->children[i]);
    }
  }
}

static void initialize(quadtree_t *qt, Line **lines, size_t n, double time_step);

quadtree_t *quadtree_make(Line **lines, size_t n, double time_step) {
  quadtree_t *qt = (quadtree_t *)malloc(sizeof(*qt));
  initialize(qt, lines, n, time_step);
  return qt;
}

inline static void
update_bounds(Vec p,
              vec_dimension *minX,
              vec_dimension *maxX,
              vec_dimension *minY,
              vec_dimension *maxY) {
  *minX = p.x < *minX ? p.x : *minX;
  *maxX = p.x > *maxX ? p.x : *maxX;
  *minY = p.y < *minY ? p.y : *minY;
  *maxY = p.y > *maxY ? p.y : *maxY;
}

static void initialize(quadtree_t *qt, Line **lines, size_t n, double time_step) {
  assert(n >= 0);

  // Base case

  if (n <= LINE_THRESHOLD) {
    qt->x_mid = 0.0;
    qt->y_mid = 0.0;
    qt->lines = (Line **)malloc(n * sizeof(Line *));
    memcpy(qt->lines, lines, n * sizeof(Line *));
    qt->num_lines = n;
    qt->children = NULL;
    qt->time_step = time_step;
    return;
  }

  // Recursive case

  // Find a point to divide into four

  vec_dimension minX = INFINITY;
  vec_dimension maxX = -INFINITY;
  vec_dimension minY = INFINITY;
  vec_dimension maxY = -INFINITY;
  for (size_t i = 0; i < n; i++) {
    update_bounds(lines[i]->p1, &minX, &maxX, &minY, &maxY);
    update_bounds(lines[i]->p2, &minX, &maxX, &minY, &maxY);
  }

  vec_dimension x_mid = (minX + maxX) / 2.0;
  vec_dimension y_mid = (minY + maxY) / 2.0;

  // Label each line with a quadrant or SELF

  size_t counts[CHILD_COUNT + 1] = {0};
  line_category_t categories[n];
  for (size_t i = 0; i < n; i++) {
    vec_dimension minX, maxX, minY, maxY;
    boundingBox(lines[i], time_step, &minX, &maxX, &minY, &maxY);
    if (minX >= x_mid) {
      if (minY >= y_mid) {
        categories[i] = NE;
        counts[NE]++;
        continue;
      } else if (maxY < y_mid) {
        categories[i] = SE;
        counts[SE]++;
        continue;
      }
    } else if (maxX < x_mid) {
      if (minY >= y_mid) {
        categories[i] = NW;
        counts[NW]++;
        continue;
      } else if (maxY < y_mid) {
        categories[i] = SW;
        counts[SW]++;
        continue;
      }
    }

    // Didn't fit into any of the four quadrants so assign it to self.
    categories[i] = SELF;
    counts[SELF]++;
  }

  // Copy the lines into new arrays for each quadrant

  assert(counts[NW] >= 0);
  assert(counts[NE] >= 0);
  assert(counts[SW] >= 0);
  assert(counts[SE] >= 0);
  assert(counts[SELF] >= 0);
  assert(counts[NW] + counts[NE] + counts[SW] + counts[SE] + counts[SELF] == n);

  Line **self_lines = (Line **)malloc(counts[SELF] * sizeof(Line *));
  Line **nw_lines   = (Line **)malloc(counts[NW]   * sizeof(Line *));
  Line **ne_lines   = (Line **)malloc(counts[NE]   * sizeof(Line *));
  Line **sw_lines   = (Line **)malloc(counts[SW]   * sizeof(Line *));
  Line **se_lines   = (Line **)malloc(counts[SE]   * sizeof(Line *));

  size_t self_i = 0, nw_i = 0, ne_i = 0, sw_i = 0, se_i = 0;

  for (size_t i = 0; i < n; i++) {
    switch (categories[i]) {
    case NW:
      assert(nw_i < counts[NW]);
      nw_lines[nw_i++] = lines[i];
      break;
    case NE:
      assert(ne_i < counts[NE]);
      ne_lines[ne_i++] = lines[i];
      break;
    case SW:
      assert(sw_i < counts[SW]);
      sw_lines[sw_i++] = lines[i];
      break;
    case SE:
      assert(se_i < counts[SE]);
      se_lines[se_i++] = lines[i];
      break;
    case SELF:
      assert(self_i < counts[SELF]);
      self_lines[self_i++] = lines[i];
      break;
    }
  }

  assert(nw_i == counts[NW]);
  assert(ne_i == counts[NE]);
  assert(sw_i == counts[SW]);
  assert(se_i == counts[SE]);
  assert(self_i == counts[SELF]);

  qt->lines     = self_lines;
  qt->num_lines = counts[SELF];

  // Allocate and make child quadtrees
  qt->children = (quadtree_t *)malloc(CHILD_COUNT * sizeof(quadtree_t));

  initialize(&qt->children[NW], nw_lines, counts[NW], time_step);
  initialize(&qt->children[NE], ne_lines, counts[NE], time_step);
  initialize(&qt->children[SW], sw_lines, counts[SW], time_step);
  initialize(&qt->children[SE], se_lines, counts[SE], time_step);

  // OK to free because `initialize` makes a copy
  free(nw_lines);
  free(ne_lines);
  free(sw_lines);
  free(se_lines);

  qt->time_step = time_step;

  assert_invariant(qt);
}

inline static void intersect_any_order(Line *l1, Line *l2, double time_step,
                                       IntersectionEventListReducer *intersection_event_list);

static void intersections_with(const quadtree_t *quadtree, Line *line, IntersectionEventListReducer *intersection_event_list);

void quadtree_intersections(const quadtree_t *const quadtree, IntersectionEventListReducer *intersection_event_list) {
  assert_invariant(quadtree);
  double time_step = quadtree->time_step;

  // Pairs of lines in this node
  cilk_for (size_t i = 0; i < quadtree->num_lines; i++) {
    cilk_for (size_t j = i + 1; j < quadtree->num_lines; j++) {
      assert (i != j);
      Line *l1 = quadtree->lines[i];
      Line *l2 = quadtree->lines[j];
      intersect_any_order(l1, l2, time_step, intersection_event_list);
    }
  }

  if (quadtree->children != NULL) {
    cilk_for (size_t i = 0; i < CHILD_COUNT; i++) {
      const quadtree_t *const child = &quadtree->children[i];
      // intersections within child node
      quadtree_intersections(child, intersection_event_list);
      // intersections between lines in this node and lines in child nodes
      cilk_for (size_t j = 0; j < quadtree->num_lines; j++) {
        Line *line = quadtree->lines[j];
        intersections_with(child, line, intersection_event_list);
      }
    }
  }
}

static void intersections_with(const quadtree_t *const quadtree, Line *const line, IntersectionEventListReducer *intersection_event_list) {
  assert_invariant(quadtree);
  double time_step = quadtree->time_step;
  for (size_t i = 0; i < quadtree->num_lines; i++) {
    Line *self_line = quadtree->lines[i];
    intersect_any_order(line, self_line, time_step, intersection_event_list);
  }

  if (quadtree->children != NULL) {
    for (size_t i = 0; i < CHILD_COUNT; i++) {
      intersections_with(&quadtree->children[i], line, intersection_event_list);
    }
  }
}

inline static void intersect_any_order(Line *l1, Line *l2, const double time_step, IntersectionEventListReducer *intersection_event_list) {
  // intersect expects compareLines(l1, l2) < 0 to be true.
  // Swap l1 and l2, if necessary.
  assert(l1 != l2);
  if (compareLines(l1, l2) >= 0) {
    Line *temp = l1;
    l1 = l2;
    l2 = temp;
  }
  IntersectionType intersection_type = intersect(l1, l2, time_step);

  if (intersection_type != NO_INTERSECTION) {
    IntersectionEventList_appendNode(&REDUCER_VIEW(*intersection_event_list), l1, l2, intersection_type);
  }
}

void quadtree_destroy(quadtree_t *quadtree) {
  assert(quadtree != NULL);
  free(quadtree->lines);
  if (quadtree->children != NULL) {
    for (int i = 0; i < CHILD_COUNT; i++) {
      quadtree_destroy(&quadtree->children[i]);
    }
    free(quadtree->children);
  }
}
